/*
   mbrot.c - Program to create a picture of a portion of the Mandelbrot set
	     for the equation Z = Z + C

	       Program by Brian Irvine
	       Version 1.10    1/1/86

	     Program performs two functions separately:

		   1.  Allows choosing an area of the complex plane to be
		       displayed by specifying the center of the picture
		       and the width of the area to be displayed, then
		       calculates points for the chosen domain and stores
		       them in a disk file.
		   2.  Reads the data points from the named disk file
		       then plots the picture on the medium-resolution
		       graphics screen in 4 colors.

   This is an implementation of the program Mandelzoom described by
   A. K. Dewdney in the August 1985 edition of Scientific American.

   Further material is available in a beautifully produced book titled
   "Frontiers of Chaos" ("Schonheit Im Chaos") by H. O. Peitgen and
   P. H. Richter of the Forschungsgruppe Komplexe Dynamik, University of
   Bremen, now available in bookstores.

   Revised 12/12/85    Version 1.1
       Added the ability to specify file containing color distribution
       for more flexible color definition.

*/

#include   "stdio.h"
#include   "math.h"


/*----- Physical attributes of IBM color screen ----------------------------*/

#define    MAXCOL  319	   /* IBM screen width and height parms */
#define    MAXROW  199	   /* Horizontal (0-319) and vertical (0-199) */
#define    PIXLRAT .6250   /* pixel ratio ( h/w = 200/320 ) */
#define    ASPECT  .7500   /* Screen aspect ratio ( h/w ) */

#define    GRMODE  4	   /* IBM medium-res graphics mode */
#define    TEXT    7	   /* IBM monochrome text mode */
#define    CTEXT   3	   /* IBM 80x25 color mode */

/*--------------------------------------------------------------------------*/

#define    MAXPTS  16000   /* size of points array */
#define    LIMIT   100	   /* maximum number of iterations */
#define    MAXSIZE 4	   /* square of the maximum allowable magnitude of Z */

#define    READ    0	   /* file mode for open() */
#define    ESC_CHAR 27
#define    F_1	   210
#define    F_2	   211

int    count [MAXPTS];	   /* This is the array which holds the iteration */
			   /* counts for each point of the area to be viewed. */
			   /* This array will be filled up four times and */
			   /* appended to the disk file each time. This is */
			   /* because the definition of the 320x200 screen */
			   /* requires 64000 points, each requiring 2 bytes */
			   /* to store the iteration count.  Increasing the */
			   /* cutoff for the iteration count increases the */
			   /* resolution of the picture generated. */

int    backgnd; 	   /* background color */
int    palete = 0;	   /* color pallette */
int    _rax,_rbx,_rcx,_rdx;/* register storage */
int    row, column;	   /* cursor row and column on the text screen */
int    grow, gcolumn;	   /* current row and column on the graphics screen */
int    cur_val; 	   /* iteration count for last point plotted */
long   top_limit;	   /* the total number of pixels on the screen */
FILE   picture; 	   /* the picture file descriptor */
FILE   colordat;	   /* the color definition file descriptor */
char   picname[10];	   /* name of the picture file */
char   colname[13];	   /* name of color definition file */
char   time[9]; 	   /* time string buffer */
char   numstr[6];	   /* temporarily holds string from coldat file */
int    colors [LIMIT];	   /* array holds color definitions */

double Creal;		   /* real part of number in center of area */
double Cimag;		   /* imaginary part of number in center of area */
double Old_Creal;	   /* save for restoring Creal */
double Old_Cimag;	   /* save for restoring Cimag */
double Width;		   /* width of area to be plotted */
double Height;		   /* Height of area to be plotted */
double dX, dY;		   /* delta increments for X and Y */
double Zreal;		   /* real component of Z */
double Zimag;		   /* imaginary component of Z */
double Size;		   /* magnitude of vector Z */

/*----- Function declarations ----------------------------------------------*/

char   next_ch(), chk_ch();
void   cls(), locate(), new_pic(), old_pic(), get_time(), process(),
       graphics_on(), graphics_off(), plot(), pallette();
int    iterate(), save(), get_vmode(), choose_color();

/*--------------------------------------------------------------------------*/

main()
{
   int	   i,j;
   char    choice;

   do  {
       cls();
       locate (0,0);
       puts("      MandelZoom -- by Brian Irvine\n");
       puts("              Version 1.10");
       puts("\n\n1. Calculate points for a picture");
       puts("\n2. Display a picture.");
       puts("\n\nEnter your choice => ");
       choice = toupper (next_ch());
       }   while ( choice != '1' && choice != '2');

       switch ( choice )
	   {
	   case    '1':
	       new_pic();
	       break;

	   case    '2':
	       old_pic();
	       break;
	   }


}


/*----- Start a new picture ------------------------------------------------*/

void   new_pic()
{
   int i,j, points;
   char    dummy;

   /* Get file name and parameters */

   puts("\n\nName of picture file => ");
   gets( picname );
   puts("\nX coordinate (real part) of center => ");
   scanf( "%F", &Creal );
   puts("\nY coordinate (imaginary part) of center => ");
   scanf( "%F", &Cimag );
   puts("\nHeight of area to plot => ");
   scanf( "%F", &Height );

   /* calculate the X and Y increments */

   Width = Height / ASPECT;
   dX = Width / (double) (MAXCOL);
   dY = Height / (double) (MAXROW);

   /* find the lower-left corner of the picture */

   Creal -= Width/2;
   Old_Creal = Creal;
   Cimag += Height/2;
   Old_Cimag = Cimag;

   /* open the file */

   cls();
   strcat(picname, ".PIC");
   if ( (picture = creat (picname, "a") ) == ERR )
       {
       puts("\n\n*** Error creating file ***\n***   Program Halted    ***\n");
       exit(1);
       }
   locate(0,0);
   puts("MandelZoom - by Brian Irvine\n");
   printf("\nCreating picture data file %s", picname);
   puts("\nThe region being calculated is:\n");
   printf("     From %1.5F, %1.5F i\n",Creal, Cimag);
   printf("     To   %1.5F, %1.5F i\n",Creal+Width, Cimag-Height);
   get_time();
   printf("\nBegun at %s", time);
   puts("\n\nCurrently working on row ");

   /* start calculating */

   points = 0;
   for ( grow = 0; grow <= MAXROW; grow++ )
       {
       locate ( 9,25 );
       printf ("%-3d", grow);
       for ( gcolumn = 0; gcolumn <= MAXCOL; gcolumn++ )
	   {
	   count[points++] = iterate();
	   if ( points == MAXPTS )
	       if ( save(points) == ERR )
		   {
		   puts("\n\nError saving data to disk.\nProgram halted.");
		   close(picture);
		   exit(1);
		   }
	       else
		   points = 0;
	   Creal = Old_Creal + ( dX * gcolumn );
	   }
       Creal = Old_Creal;
       Cimag = Old_Cimag - ( dY * grow );
       }

   /* if top_limit is not a multiple of MAXPTS, */
   /* then write the remaining values to the file */

   if ( points )
       save(points);
   close(picture);
   get_time();
   printf("\nFinished at: %s", time );
}


/*----- Process an existing file -------------------------------------------*/

void   old_pic()
{
   char    ch;
   int	   number, i;

   if ( get_vmode() == TEXT )
       {
       puts("\nSorry, this part requires a graphics adapter.\nProgram halted.\n\n");
       exit(1);
       }
   puts("\n\nName of color definition file => ");
   gets( colname );
   if ( ( colordat = fopen ( colname, "r" ) ) == NULL )
       {
       printf ( "\nCan't find the file %s\nProgram halted.\n", colname);
       exit(1);
       }
   for ( i = 0; i < LIMIT; i++ )
       {
       fscanf( colordat, "%5s", numstr);
       colors [i] = atoi ( numstr );
       }
   puts("\n\nName of picture file => ");

   gets( picname );
   strcat(picname, ".PIC");

   /* open the file */

   if ( (picture = open(picname, READ) ) == ERR )
       {
       puts("\n\nCan't find that file.\nProgram halted.\n\n");
       exit(1);
       }

   /* get into graphics mode */

   graphics_on();
   pallette ( palete );
   grow = 0;  gcolumn = 0;

   /* Repeatedly read in MAXPTS integers and plot the points. */
   /* if less than MAXPTS left, read in the remainder */

   top_limit = (long)( MAXCOL + 1 ) * (long)( MAXROW + 1 );
   while ( top_limit )
       {
       top_limit -= MAXPTS;
       if ( top_limit >= 0 )
	   number = MAXPTS;
       else
	   {
	   number = top_limit + MAXPTS;
	   top_limit = 0;
	   }
       read ( picture, count, number * sizeof (int) );
       process(number);
       }

   /* wait for the Esc key before returning to DOS */
   /* allow playing around with background and pallette */

   close ( picture );
   while ( (ch = next_ch()) != ESC_CHAR )
       {
       switch (ch)
	   {
	   case F_1:
	       toggle_pal();
	       break;

	   case F_2:
	       inc_bkg();
	       break;

	   default:
	       break;

	   }
       }
   graphics_off();

}



/*----- Toggle color pallette ----------------------------------------------*/

void   toggle_pal()
{
   palete = ( palete == 0 ) ? 1 : 0;
   pallette ( palete );
}



/*----- Increment background color -----------------------------------------*/

void   inc_bkg()
{
   backgnd = ( ++backgnd > 15 ) ? 0 : backgnd;
   set_backgnd ();
}



/*----- Process points -----------------------------------------------------*/

void   process(number)
int    number;
{
   int	   i, color;

   for ( i=0; i < number; i++ )
       {
       color = choose_color(i);
       plot ( grow, gcolumn, color );
       gcolumn++;
       if ( gcolumn > MAXCOL )
	   {
	   gcolumn = 0;
	   grow++;
	   }
       }

}



/*----- Choose point color -------------------------------------------------*/
/*									    */
/*     This routine is passed the index into the count[] array for the	    */
/*     next point to be plotted, and returns the color to plot there.	    */
/*									    */
/*--------------------------------------------------------------------------*/

int    choose_color(point)
int    point;
{

   return ( colors [ count [point] ] );

}


/*----- Calculate ----------------------------------------------------------*/
/*									    */
/*     Repeatedly calculate Z = Z + C until the iteration count reaches    */
/*     the maximum or until the magnitude of Z exceeds a threshold value.   */
/*     Returns the number of iterations.				    */
/*									    */
/*--------------------------------------------------------------------------*/


int    iterate()
{
   int	   count;
   char    ch;
   double  RealZ;

   Zreal = Zimag = 0;
   for ( count = 1; count <= LIMIT; count++ )
       {
       if ( (ch = chk_ch()) == ESC_CHAR )
	   {
	   close (picture);
	   exit(0);
	   }
       RealZ = Zreal;
       Zreal = RealZ * RealZ - Zimag * Zimag + Creal;
       Zimag = 2.0 * RealZ * Zimag + Cimag;
       Size = Zreal * Zreal + Zimag * Zimag;
       if ( Size > MAXSIZE )
	   return ( count );
       }
   return (LIMIT);

}


/*----- Save the array count[] to disk -------------------------------------*/

int    save(number)
int    number;
{
   return ( write ( picture, count, number * sizeof(int) ) );

}


/****************************************************************************/
/*     Functions specific to DeSmet compiler				    */


/*----- Clear the screen ---------------------------------------------------*/

void   cls()
{
   scr_clr();

}


/*----- Set cursor position ------------------------------------------------*/

void   locate ( row, col )
int    row, col;
{
   scr_rowcol( row, col );

}


/*----- Wait for next character from keyboard ------------------------------*/

char   next_ch()
{
   char    ch;

   ch = scr_ci();
   return ch;
}


/*----- Check for character and return with it if present ------------------*/

char   chk_ch()
{
   char    ch;
   ch = scr_csts();
   return ch;

}


/*----- Get the time into the string time[] --------------------------------*/

void   get_time()
{
   times(time);
}


/*----- Get the current video mode -----------------------------------------*/

int    get_vmode()
{
   _rax = 0x0F00;
   _doint ( 0x10 );
   return ( _rax & 0x00FF );

}


/*----- Plot a point on the graphics screen --------------------------------*/

void   plot( y,x,c )
int    y,x,c;
{
   _rax = 0x0C00 + c;
   _rdx = y;
   _rcx = x;
   _doint( 0x10 );

}


/*----- Set graphics mode on -----------------------------------------------*/

void   graphics_on()
{
   scr_setmode( GRMODE );

}

/*----- Set graphics mode off ----------------------------------------------*/

void   graphics_off()
{
   scr_setmode( CTEXT );

}

/*--------------------------------------------------------------------------*/

void   pallette ( set )
int    set;
{
   _rax = 0x0B00;
   _rbx = 0x0100 + set;
   _doint( 0x10 );
}


/*----- Set Background Color -----------------------------------------------*/

void   set_backgnd ( color )
int    color;
{
   _rax = 0x0B00;
   _rbx = backgnd;
   _doint( 0x10 );
}

